﻿/// <summary>Converters for discriminated union types</summary>
module MyWebLog.Converters

open MyWebLog
open System

/// <summary>JSON.NET converters for discriminated union types</summary>
module Json =

    open Newtonsoft.Json

    /// <summary>Converter for the <see cref="CategoryId" /> type</summary>
    type CategoryIdConverter() =
        inherit JsonConverter<CategoryId>()
        override _.WriteJson(writer: JsonWriter, value: CategoryId, _: JsonSerializer) =
            writer.WriteValue(string value)
        override _.ReadJson(reader: JsonReader, _: Type, _: CategoryId, _: bool, _: JsonSerializer) =
            (string >> CategoryId) reader.Value

    /// <summary>Converter for the <see cref="CommentId" /> type</summary>
    type CommentIdConverter() =
        inherit JsonConverter<CommentId>()
        override _.WriteJson(writer: JsonWriter, value: CommentId, _: JsonSerializer) =
            writer.WriteValue(string value)
        override _.ReadJson(reader: JsonReader, _: Type, _: CommentId, _: bool, _: JsonSerializer) =
            (string >> CommentId) reader.Value

    /// <summary>Converter for the <see cref="CommentStatus" /> type</summary>
    type CommentStatusConverter() =
        inherit JsonConverter<CommentStatus>()
        override _.WriteJson(writer: JsonWriter, value: CommentStatus, _: JsonSerializer) =
            writer.WriteValue(string value)
        override _.ReadJson(reader: JsonReader, _: Type, _: CommentStatus, _: bool, _: JsonSerializer) =
            (string >> CommentStatus.Parse) reader.Value

    /// <summary>Converter for the <see cref="CustomFeedId" /> type</summary>
    type CustomFeedIdConverter() =
        inherit JsonConverter<CustomFeedId>()
        override _.WriteJson(writer: JsonWriter, value: CustomFeedId, _: JsonSerializer) =
            writer.WriteValue(string value)
        override _.ReadJson(reader: JsonReader, _: Type, _: CustomFeedId, _: bool, _: JsonSerializer) =
            (string >> CustomFeedId) reader.Value

    /// <summary>Converter for the <see cref="CustomFeedSource" /> type</summary>
    type CustomFeedSourceConverter() =
        inherit JsonConverter<CustomFeedSource>()
        override _.WriteJson(writer: JsonWriter, value: CustomFeedSource, _: JsonSerializer) =
            writer.WriteValue(string value)
        override _.ReadJson(reader: JsonReader, _: Type, _: CustomFeedSource, _: bool, _: JsonSerializer) =
            (string >> CustomFeedSource.Parse) reader.Value

    /// <summary>Converter for the <see cref="ExplicitRating" /> type</summary>
    type ExplicitRatingConverter() =
        inherit JsonConverter<ExplicitRating>()
        override _.WriteJson(writer: JsonWriter, value: ExplicitRating, _: JsonSerializer) =
            writer.WriteValue(string value)
        override _.ReadJson(reader: JsonReader, _: Type, _: ExplicitRating, _: bool, _: JsonSerializer) =
            (string >> ExplicitRating.Parse) reader.Value

    /// <summary>Converter for the <see cref="MarkupText" /> type</summary>
    type MarkupTextConverter() =
        inherit JsonConverter<MarkupText>()
        override _.WriteJson(writer: JsonWriter, value: MarkupText, _: JsonSerializer) =
            writer.WriteValue(string value)
        override _.ReadJson(reader: JsonReader, _: Type, _: MarkupText, _: bool, _: JsonSerializer) =
            (string >> MarkupText.Parse) reader.Value

    /// <summary>Converter for the <see cref="Permalink" /> type</summary>
    type PermalinkConverter() =
        inherit JsonConverter<Permalink>()
        override _.WriteJson(writer: JsonWriter, value: Permalink, _: JsonSerializer) =
            writer.WriteValue(string value)
        override _.ReadJson(reader: JsonReader, _: Type, _: Permalink, _: bool, _: JsonSerializer) =
            (string >> Permalink) reader.Value

    /// <summary>Converter for the <see cref="PageId" /> type</summary>
    type PageIdConverter() =
        inherit JsonConverter<PageId>()
        override _.WriteJson(writer: JsonWriter, value: PageId, _: JsonSerializer) =
            writer.WriteValue(string value)
        override _.ReadJson(reader: JsonReader, _: Type, _: PageId, _: bool, _: JsonSerializer) =
            (string >> PageId) reader.Value

    /// <summary>Converter for the <see cref="PodcastMedium" /> type</summary>
    type PodcastMediumConverter() =
        inherit JsonConverter<PodcastMedium>()
        override _.WriteJson(writer: JsonWriter, value: PodcastMedium, _: JsonSerializer) =
            writer.WriteValue(string value)
        override _.ReadJson(reader: JsonReader, _: Type, _: PodcastMedium, _: bool, _: JsonSerializer) =
            (string >> PodcastMedium.Parse) reader.Value

    /// <summary>Converter for the <see cref="PostId" /> type</summary>
    type PostIdConverter() =
        inherit JsonConverter<PostId>()
        override _.WriteJson(writer: JsonWriter, value: PostId, _: JsonSerializer) =
            writer.WriteValue(string value)
        override _.ReadJson(reader: JsonReader, _: Type, _: PostId, _: bool, _: JsonSerializer) =
            (string >> PostId) reader.Value

    /// <summary>Converter for the <see cref="TagMapId" /> type</summary>
    type TagMapIdConverter() =
        inherit JsonConverter<TagMapId>()
        override _.WriteJson(writer: JsonWriter, value: TagMapId, _: JsonSerializer) =
            writer.WriteValue(string value)
        override _.ReadJson(reader: JsonReader, _: Type, _: TagMapId, _: bool, _: JsonSerializer) =
            (string >> TagMapId) reader.Value

    /// <summary>Converter for the <see cref="ThemeAssetId" /> type</summary>
    type ThemeAssetIdConverter() =
        inherit JsonConverter<ThemeAssetId>()
        override _.WriteJson(writer: JsonWriter, value: ThemeAssetId, _: JsonSerializer) =
            writer.WriteValue(string value)
        override _.ReadJson(reader: JsonReader, _: Type, _: ThemeAssetId, _: bool, _: JsonSerializer) =
            (string >> ThemeAssetId.Parse) reader.Value

    /// <summary>Converter for the <see cref="ThemeId" /> type</summary>
    type ThemeIdConverter() =
        inherit JsonConverter<ThemeId>()
        override _.WriteJson(writer: JsonWriter, value: ThemeId, _: JsonSerializer) =
            writer.WriteValue(string value)
        override _.ReadJson(reader: JsonReader, _: Type, _: ThemeId, _: bool, _: JsonSerializer) =
            (string >> ThemeId) reader.Value

    /// <summary>Converter for the <see cref="UploadId" /> type</summary>
    type UploadIdConverter() =
        inherit JsonConverter<UploadId>()
        override _.WriteJson(writer: JsonWriter, value: UploadId, _: JsonSerializer) =
            writer.WriteValue(string value)
        override _.ReadJson(reader: JsonReader, _: Type, _: UploadId, _: bool, _: JsonSerializer) =
            (string >> UploadId) reader.Value

    /// <summary>Converter for the <see cref="WebLogId" /> type</summary>
    type WebLogIdConverter() =
        inherit JsonConverter<WebLogId>()
        override _.WriteJson(writer: JsonWriter, value: WebLogId, _: JsonSerializer) =
            writer.WriteValue(string value)
        override _.ReadJson(reader: JsonReader, _: Type, _: WebLogId, _: bool, _: JsonSerializer) =
            (string >> WebLogId) reader.Value

    /// <summary>Converter for the <see cref="WebLogUserId" /> type</summary>
    type WebLogUserIdConverter() =
        inherit JsonConverter<WebLogUserId> ()
        override _.WriteJson(writer: JsonWriter, value: WebLogUserId, _: JsonSerializer) =
            writer.WriteValue(string value)
        override _.ReadJson(reader: JsonReader, _: Type, _: WebLogUserId, _: bool, _: JsonSerializer) =
            (string >> WebLogUserId) reader.Value

    open Microsoft.FSharpLu.Json
    open NodaTime
    open NodaTime.Serialization.JsonNet

    /// <summary>Configure a serializer to use these converters (and other settings)</summary>
    let configure (ser: JsonSerializer) =
        // Our converters
        [ CategoryIdConverter() :> JsonConverter
          CommentIdConverter()
          CommentStatusConverter()
          CustomFeedIdConverter()
          CustomFeedSourceConverter()
          ExplicitRatingConverter()
          MarkupTextConverter()
          PermalinkConverter()
          PageIdConverter()
          PodcastMediumConverter()
          PostIdConverter()
          TagMapIdConverter()
          ThemeAssetIdConverter()
          ThemeIdConverter()
          UploadIdConverter()
          WebLogIdConverter()
          WebLogUserIdConverter() ]
        |> List.iter ser.Converters.Add
        // NodaTime
        let _ = ser.ConfigureForNodaTime DateTimeZoneProviders.Tzdb
        // Handles DUs with no associated data, as well as option fields
        ser.Converters.Add(CompactUnionJsonConverter())
        ser.NullValueHandling     <- NullValueHandling.Ignore
        ser.MissingMemberHandling <- MissingMemberHandling.Ignore
        ser

    /// Serializer settings extracted from a JsonSerializer (a property sure would be nice...)
    let mutable private serializerSettings : JsonSerializerSettings option = None

    /// <summary>Extract settings from the serializer to be used in <c>JsonConvert</c> calls</summary>
    /// <param name="ser">The serializer from which settings will be extracted if required</param>
    /// <returns>The serializer settings to use for <c>JsonConvert</c> calls</returns>
    let settings (ser: JsonSerializer) =
        if Option.isNone serializerSettings then
            serializerSettings <- JsonSerializerSettings (
                ConstructorHandling            = ser.ConstructorHandling,
                ContractResolver               = ser.ContractResolver,
                Converters                     = ser.Converters,
                DefaultValueHandling           = ser.DefaultValueHandling,
                DateFormatHandling             = ser.DateFormatHandling,
                DateParseHandling              = ser.DateParseHandling,
                MetadataPropertyHandling       = ser.MetadataPropertyHandling,
                MissingMemberHandling          = ser.MissingMemberHandling,
                NullValueHandling              = ser.NullValueHandling,
                ObjectCreationHandling         = ser.ObjectCreationHandling,
                ReferenceLoopHandling          = ser.ReferenceLoopHandling,
                SerializationBinder            = ser.SerializationBinder,
                TraceWriter                    = ser.TraceWriter,
                TypeNameAssemblyFormatHandling = ser.TypeNameAssemblyFormatHandling,
                TypeNameHandling               = ser.TypeNameHandling)
                |> Some
        serializerSettings.Value
